"""Config loading methods and helpers.

This is designed to house the main functions which are exposed by the
overall config module. There is some caching in this module, which
is designed around caching the configuration loaded at *specific paths*
rather than the individual file caching in the `file` module.
"""

from __future__ import annotations

import logging
import os
import os.path
import sys
import threading
from functools import cache
from importlib.resources import files
from pathlib import Path
from typing import (
    Optional,
)

import platformdirs
import platformdirs.macos
import platformdirs.unix

from sqlfluff.core.config.file import (
    load_config_file_as_dict,
    load_config_string_as_dict,
)
from sqlfluff.core.errors import SQLFluffUserError
from sqlfluff.core.helpers.dict import nested_combine
from sqlfluff.core.helpers.file import iter_intermediate_paths
from sqlfluff.core.types import ConfigMappingType

# Instantiate the config logger
config_logger = logging.getLogger("sqlfluff.config")

# Lock for thread-safe environment variable manipulation in
# _get_user_config_dir_path. Needed for free-threaded Python (3.14+).
_env_lock = threading.Lock()

global_loader = None
""":obj:`ConfigLoader`: A variable to hold the single module loader when loaded.

We define a global loader, so that between calls to load config, we
can still cache appropriately
"""


def _get_user_config_dir_path(sys_platform: str) -> str:
    """Get the user config dir for this system.

    Args:
        sys_platform (str): The result of ``sys.platform()``. Provided
            as an argument here for ease of testing. In normal usage
            it should only be  called with ``sys.platform()``. This
            argument only applies to switching between linux and macos.
            Win32 detection still uses the underlying ``sys.platform()``
            methods.
    """
    appname = "sqlfluff"
    appauthor = "sqlfluff"

    # First try the default SQLFluff specific cross-platform config path.
    cross_platform_path = os.path.expanduser("~/.config/sqlfluff")
    if os.path.exists(cross_platform_path):
        return cross_platform_path

    # Then try the platform specific paths, for MacOS, we check
    # the unix variant first to preferentially use the XDG config path if set.
    # https://github.com/sqlfluff/sqlfluff/issues/889
    if sys_platform == "darwin":
        unix_config_path = platformdirs.unix.Unix(
            appname=appname, appauthor=appauthor
        ).user_config_dir
        if os.path.exists(os.path.expanduser(unix_config_path)):
            return unix_config_path
        # Technically we could just delegate to the generic `user_config_dir`
        # method, but for testing it's convenient to explicitly call the macos
        # methods here.
        # NOTE: On platformdirs >= 4.6.0, MacOS also respects XDG_CONFIG_HOME.
        # Temporarily unset it so we get the native macOS path when the XDG
        # path doesn't exist. Use a lock for thread safety (free-threaded
        # Python 3.14+).
        with _env_lock:
            xdg_config_home = os.environ.pop("XDG_CONFIG_HOME", None)
            try:
                return platformdirs.macos.MacOS(
                    appname=appname, appauthor=appauthor
                ).user_config_dir
            finally:
                if xdg_config_home is not None:
                    os.environ["XDG_CONFIG_HOME"] = xdg_config_home
    # NOTE: We could delegate to the generic `user_config_dir` method here,
    # but for testing it's convenient to explicitly call the linux methods.
    elif sys_platform == "linux":
        return platformdirs.unix.Unix(
            appname=appname, appauthor=appauthor
        ).user_config_dir
    # Defer to the self-detecting paths.
    # NOTE: On Windows this means that the `sys_platform` argument is not
    # applied.
    return platformdirs.user_config_dir(appname, appauthor)


def load_config_file(
    file_dir: str, file_name: str, configs: Optional[ConfigMappingType] = None
) -> ConfigMappingType:
    """Load a config file from the filesystem.

    Args:
        file_dir (str): The path to the location of file to be loaded.
            This should be a reference to the directory *only* and not
            include the filename itself. Any paths in the loaded file
            are resolved relative to this location.
        file_name (str): The filename of the file to be loaded. If the
            filename is ``pyproject.toml`` then the file is loaded in
            ``toml`` format, but otherwise is assumed to be in ``ini``
            format (as per ``.sqlfluff``).
        configs (ConfigMappingType, optional): A base set of configs to
            merge the loaded configs onto. If not provided, the result
            will contain only the values loaded from the string.

    Returns:
        :obj:`ConfigMappingType`: A nested dictionary of config values.
    """
    file_path = os.path.join(file_dir, file_name)
    raw_config = load_config_file_as_dict(file_path)
    # We always run `nested_combine()` because it has the side effect
    # of making a copy of the objects provided. This prevents us
    # from editing items which also sit within the cache.
    return nested_combine(configs or {}, raw_config)


def load_config_resource(package: str, file_name: str) -> ConfigMappingType:
    """Load a config resource from a python package.

    Args:
        package (str): The name of the python package to load the resource
            from.
        file_name (str): The filename of the file to be loaded. If the
            filename is ``pyproject.toml`` then the file is loaded in
            ``toml`` format, but otherwise is assumed to be in ``ini``
            format (as per ``.sqlfluff``).

    Returns:
        :obj:`ConfigMappingType`: A nested dictionary of config values.

    This is primarily used when loading configuration bundled with a
    SQLFluff plugin, or to load the default config for SQLFluff itself.
    By loading config from the package directly we avoid some of the
    path resolution which otherwise occurs. This is also more compatible
    with ``mypyc`` because it avoids the use of the ``__file__`` attribute
    to find the default config.

    Any paths found in the loaded config are resolved relative
    to ``os.getcwd()``.

    For more information about resource loading, see the docs for importlib:
    https://docs.python.org/3/library/importlib.resources.html
    """
    config_string = files(package).joinpath(file_name).read_text()
    # NOTE: load_config_string_as_dict is cached.
    return load_config_string_as_dict(
        config_string,
        os.getcwd(),
        logging_reference=f"<resource {package}.{file_name}>",
    )


def load_config_string(
    config_string: str,
    configs: Optional[ConfigMappingType] = None,
    working_path: Optional[str] = None,
) -> ConfigMappingType:
    """Load a config from a string in ini format.

    Args:
        config_string (str): The raw config file as a string. The content
            is assumed to be in the ``.ini`` format of a ``.sqlfluff``
            file (i.e. not in ``.toml`` format).
        configs (ConfigMappingType, optional): A base set of configs to
            merge the loaded configs onto. If not provided, the result
            will contain only the values loaded from the string.
        working_path (str, optional): The working path to use for the
            resolution of any paths specified in the config. If not provided
            then ``os.getcwd()`` is used as a default.

    Returns:
        :obj:`ConfigMappingType`: A nested dictionary of config values.
    """
    filepath = working_path or os.getcwd()
    raw_config = load_config_string_as_dict(
        config_string, filepath, logging_reference="<config string>"
    )
    # We always run `nested_combine()` because it has the side effect
    # of making a copy of the objects provided. This prevents us
    # from editing items which also sit within the cache.
    return nested_combine(configs or {}, raw_config)


@cache
def load_config_at_path(path: str) -> ConfigMappingType:
    """Load config files at a given path.

    Args:
        path (str): The directory to search for config files.

    Returns:
        :obj:`ConfigMappingType`: A nested dictionary of config values.

    This function will search for all valid config files at the given
    path, load any found and combine them into a config mapping. If
    multiple valid files are found, they are resolved in priority order,
    where ``pyproject.toml`` is given the highest precedence, followed
    by ``.sqlfluff``, ``pep8.ini``, ``tox.ini`` and finally ``setup.cfg``.

    By accepting only a path string, we enable efficient caching of
    results, such that configuration can be reused between files without
    reloading the information from disk.
    """
    # The potential filenames we would look for at this path.
    # NB: later in this list overwrites earlier
    filename_options = [
        "setup.cfg",
        "tox.ini",
        "pep8.ini",
        ".sqlfluff",
        "pyproject.toml",
    ]

    configs: ConfigMappingType = {}

    if os.path.isdir(path):
        p = path
    else:
        p = os.path.dirname(path)

    # Expand `~` once and reuse the expanded value for listing, checking and
    # loading. Otherwise the existence checks (which expanded the path) could
    # succeed while `load_config_file()` later fails to open the unexpanded path.
    p = os.path.expanduser(p)

    d = os.listdir(p)
    # iterate this way round to make sure things overwrite is the right direction.
    # NOTE: The `configs` variable is passed back in at each stage.
    for fname in filename_options:
        # Ignore any entries which aren't files. In particular this guards
        # against a *directory* sharing the name of a config file (e.g. a
        # directory named `.sqlfluff`), which would otherwise be opened as a
        # file and raise an `IsADirectoryError`.
        # https://github.com/sqlfluff/sqlfluff/issues/6617
        if fname in d and os.path.isfile(os.path.join(p, fname)):
            configs = load_config_file(p, fname, configs=configs)

    return configs


def _load_user_appdir_config() -> ConfigMappingType:
    """Load the config from the user's OS specific appdir config directory."""
    user_config_dir_path = _get_user_config_dir_path(sys.platform)
    if os.path.exists(user_config_dir_path):
        return load_config_at_path(user_config_dir_path)
    else:
        return {}


def load_config_up_to_path(
    path: str,
    extra_config_path: Optional[str] = None,
    ignore_local_config: bool = False,
) -> ConfigMappingType:
    """Loads a selection of config files from both the path and its parent paths.

    Args:
        path (str): The directory which is the target of the search. Config
            files in subdirectories will not be loaded by this method, but
            valid config files between this path and the current working
            path will.
        extra_config_path (str, optional): An additional path to load config
            from. This path is not used in iterating through intermediate
            paths, and is loaded last (taking the highest precedence in
            combining the loaded configs).
        ignore_local_config (bool, optional, defaults to False): If set to
            True, this skips loading configuration from the user home
            directory (``~``) or ``appdir`` path.

    Returns:
        :obj:`ConfigMappingType`: A nested dictionary of config values.

    We layer each of the configs on top of each other, starting with any home
    or user configs (e.g. in ``appdir`` or home (``~``)), then any local
    project configuration and then any explicitly specified config paths.
    """
    # 1) AppDir & Home config
    if not ignore_local_config:
        user_appdir_config = _load_user_appdir_config()
        user_config = load_config_at_path(os.path.expanduser("~"))
    else:
        user_config, user_appdir_config = {}, {}

    # 3) Local project config
    parent_config_stack = []
    config_stack = []
    if not ignore_local_config:
        # Finding all paths between here and the home
        # directory. We could start at the root of the filesystem,
        # but depending on the user's setup, this might result in
        # permissions errors.
        parent_config_paths = list(
            iter_intermediate_paths(
                Path(path).absolute(), Path(os.path.expanduser("~"))
            )
        )
        # Stripping off the home directory and the current working
        # directory, since they are both covered by other code
        # here
        parent_config_paths = parent_config_paths[1:-1]
        parent_config_stack = [
            load_config_at_path(str(p.resolve())) for p in list(parent_config_paths)
        ]
        # Resolve paths to ensure caching is accurate.
        config_paths = iter_intermediate_paths(Path(path).absolute(), Path.cwd())
        config_stack = [load_config_at_path(str(p.resolve())) for p in config_paths]

    # 4) Extra config paths.
    # When calling `load_config_file_as_dict` we resolve the path first so that caching
    # is more efficient.
    extra_config = {}
    if extra_config_path:
        # Expand `~` once so the directory check and resolution stay consistent.
        # `Path.resolve()` does not expand `~`, so without this a quoted path
        # like `--config "~/.sqlfluff"` would skip the directory check and be
        # reported as not-found instead of loaded (or flagged as a directory).
        expanded_config_path = os.path.expanduser(extra_config_path)
        # A directory (e.g. one named `.sqlfluff`) is not a valid config file
        # and would otherwise raise an `IsADirectoryError` when opened. Surface
        # a clean user error instead, consistent with the not-found case.
        # https://github.com/sqlfluff/sqlfluff/issues/6617
        if os.path.isdir(expanded_config_path):
            raise SQLFluffUserError(
                f"Extra config path '{extra_config_path}' is a directory, "
                "not a config file."
            )
        try:
            extra_config = load_config_file_as_dict(
                str(Path(expanded_config_path).resolve())
            )
        except FileNotFoundError:
            raise SQLFluffUserError(
                f"Extra config path '{extra_config_path}' does not exist."
            )

    return nested_combine(
        user_appdir_config,
        user_config,
        *parent_config_stack,
        *config_stack,
        extra_config,
    )


class ConfigLoader:
    """The class for loading config files.

    NOTE: Deprecated class maintained because it was in our example
    plugin for a long while. Remove once this warning has been live for
    an appropriate amount of time.
    """

    def __init__(self) -> None:  # pragma: no cover
        config_logger.warning(
            "ConfigLoader is deprecated, and no longer necessary. "
            "Please update your plugin to use the config loading functions directly "
            "to remove this message."
        )

    @classmethod
    def get_global(cls) -> ConfigLoader:  # pragma: no cover
        """Get the singleton loader."""
        config_logger.warning(
            "ConfigLoader.get_global() is deprecated, and no longer necessary. "
            "Please update your plugin to use the config loading functions directly "
            "to remove this message."
        )
        return cls()

    def load_config_resource(
        self, package: str, file_name: str
    ) -> ConfigMappingType:  # pragma: no cover
        """Load a config resource.

        NOTE: Deprecated classmethod maintained because it was in our example
        plugin for a long while. Remove once this warning has been live for
        an appropriate amount of time.
        """
        config_logger.warning(
            "ConfigLoader.load_config_resource() is deprecated. Please update "
            "your plugin to call sqlfluff.core.config.loader.load_config_resource() "
            "directly to remove this message."
        )
        return load_config_resource(package, file_name)



r"""A list of all SQL key words.

https://learn.microsoft.com/en-us/sql/t-sql/language-elements/reserved-keywords-transact-sql?view=sql-server-ver17

Run the script in a browser console to extract all reserved keywords:

```js
(function () {
    const xpathResult = document.evaluate(
        '//div[@class=\'column\']/p[not(descendant::strong)]',
        document,
        null,
        XPathResult.ORDERED_NODE_SNAPSHOT_TYPE,
        null
    );
    const list = new Set();

    for (let index = 0; index < xpathResult.snapshotLength; ++index) {
        const node = xpathResult.snapshotItem(index);
        list.add(node.textContent.trim());
    }
    console.log([...list].sort().map(value => `    "${value}"`).join(',\n'));
})();
```

Be careful, some keywords are present in `UNRESERVED_KEYWORDS`.
"""

RESERVED_KEYWORDS = [
    "ADD",
    "ALL",
    "ALTER",
    "AND",
    "ANY",
    "AS",
    "ASC",
    "AUTHORIZATION",
    "BACKUP",
    "BEGIN",
    "BETWEEN",
    "BREAK",
    "BROWSE",
    "BULK",
    "BY",
    "CASCADE",
    "CASE",
    "CHECK",
    "CHECKPOINT",
    "CLOSE",
    "CLUSTERED",
    "COALESCE",
    "COLLATE",
    "COLUMN",
    "COMMIT",
    "COMPUTE",
    "CONSTRAINT",
    "CONTAINS",
    "CONTAINSTABLE",
    "CONTINUE",
    "CONVERT",
    "CREATE",
    "CROSS",
    "CURRENT",
    "CURRENT_DATE",
    "CURRENT_TIME",
    "CURRENT_TIMESTAMP",
    "CURRENT_USER",
    "CURSOR",
    "DATABASE",
    "DBCC",
    "DEALLOCATE",
    "DECLARE",
    "DEFAULT",
    "DEFINITION",
    "DELETE",
    "DENY",
    "DESC",
    "DISK",
    "DISTINCT",
    "DISTRIBUTED",
    "DOUBLE",
    "DROP",
    "DUMP",
    "ELSE",
    "END",
    "ERRLVL",
    "ESCAPE",
    "EXCEPT",
    "EXEC",
    "EXECUTE",
    "EXISTS",
    "EXIT",
    "EXTERNAL",
    "FETCH",
    "FILE",
    "FILLFACTOR",
    "FOR",
    "FOREIGN",
    "FREETEXT",
    "FREETEXTTABLE",
    "FROM",
    "FULL",
    "FUNCTION",
    "GO",
    "GOTO",
    "GRANT",
    "GROUP",
    "HAVING",
    "HOLDLOCK",
    "IDENTITY",
    "IDENTITY_INSERT",
    "IDENTITYCOL",
    "IF",
    "IN",
    "INDEX",
    "INNER",
    "INSERT",
    "INTERSECT",
    "INTO",
    "IS",
    "JOIN",
    "KEY",
    "KILL",
    "LEFT",
    "LIKE",
    "LINENO",
    "LOAD",
    "MERGE",
    "NATIONAL",
    "NOCHECK",
    "NONCLUSTERED",
    "NOT",
    "NULL",
    "NULLIF",
    "OF",
    "OFF",
    "OFFSETS",
    "ON",
    "OPEN",
    "OPENDATASOURCE",
    "OPENQUERY",
    "OPENROWSET",
    "OPENXML",
    "OPTION",
    "OR",
    "ORDER",
    "OUTER",
    "OVER",
    "PERCENT",
    "PIVOT",
    "PLAN",
    "PRECISION",
    "PRIMARY",
    "PRINT",
    "PROC",
    "PROCEDURE",
    "PUBLIC",
    "RAISERROR",
    "READ",
    "READTEXT",
    "RECONFIGURE",
    "REFERENCES",
    "REPLICATION",
    "RESTORE",
    "RESTRICT",
    "RETURN",
    "REVERT",
    "REVOKE",
    "RIGHT",
    "ROLLBACK",
    "ROWCOUNT",
    "ROWGUIDCOL",
    "RULE",
    "SAVE",
    "SCHEMA",
    "SECURITYAUDIT",
    "SELECT",
    "SEMANTICKEYPHRASETABLE",
    "SEMANTICSIMILARITYDETAILSTABLE",
    "SEMANTICSIMILARITYTABLE",
    "SESSION_USER",
    "SET",
    "SETUSER",
    "SHUTDOWN",
    "SOME",
    "STATISTICS",
    "SYSTEM_USER",
    "TABLE",
    "TABLESAMPLE",
    "TEXTSIZE",
    "THEN",
    "TO",
    "TOP",
    "TRAN",
    "TRANSACTION",
    "TRIGGER",
    "TRUNCATE",
    "TRY_CONVERT",
    "TSEQUAL",
    "UNION",
    "UNIQUE",
    "UNPIVOT",
    "UPDATE",
    "UPDATETEXT",
    "USE",
    "USER",
    "VALUES",
    "VARYING",
    "VIEW",
    "WAITFOR",
    "WHEN",
    "WHERE",
    "WHILE",
    "WITH",
    "WITHIN GROUP",
    "WRITETEXT",
]

# Future reserved keywords extracted from the documentation
# NOTE: These are marked as "future reserved" by Microsoft but are NOT currently
# reserved in T-SQL. They can be used as identifiers without quoting.
# The dialect implementation only blocks RESERVED_KEYWORDS in NakedIdentifierSegment,
# allowing these future reserved keywords to be used freely as identifiers.
# https://learn.microsoft.com/en-us/sql/t-sql/language-elements/reserved-keywords-transact-sql?view=sql-server-ver17
FUTURE_RESERVED_KEYWORDS = [
    "ABSOLUTE",
    "ACTION",
    "ADMIN",
    "AFTER",
    "AGGREGATE",
    "ALIAS",
    "ALLOCATE",
    "ARE",
    "ARRAY",
    "ASENSITIVE",
    "ASSERTION",
    "ASYMMETRIC",
    "AT",
    "ATOMIC",
    "BEFORE",
    "BINARY",
    "BIT",
    "BLOB",
    "BOOLEAN",
    "BOTH",
    "BREADTH",
    "CALL",
    "CALLED",
    "CARDINALITY",
    "CASCADED",
    "CAST",
    "CATALOG",
    "CHAR",
    "CHARACTER",
    "CLASS",
    "CLOB",
    "COLLATION",
    "COLLECT",
    "COMPLETION",
    "CONDITION",
    "CONNECT",
    "CONNECTION",
    "CONSTRAINTS",
    "CONSTRUCTOR",
    "CORR",
    "CORRESPONDING",
    "COVAR_POP",
    "COVAR_SAMP",
    "CUBE",
    "CUME_DIST",
    "CURRENT_CATALOG",
    "CURRENT_DEFAULT_TRANSFORM_GROUP",
    "CURRENT_PATH",
    "CURRENT_ROLE",
    "CURRENT_SCHEMA",
    "CURRENT_TRANSFORM_GROUP_FOR_TYPE",
    "CYCLE",
    "DATA",
    "DATE",
    "DAY",
    "DEC",
    "DECIMAL",
    "DEFERRABLE",
    "DEFERRED",
    "DEPTH",
    "DEREF",
    "DESCRIBE",
    "DESCRIPTOR",
    "DESTROY",
    "DESTRUCTOR",
    "DETERMINISTIC",
    "DIAGNOSTICS",
    "DICTIONARY",
    "DISCONNECT",
    "DOMAIN",
    "DYNAMIC",
    "EACH",
    "ELEMENT",
    "END-EXEC",
    "EQUALS",
    "EVERY",
    "EXCEPTION",
    "FALSE",
    "FILTER",
    "FIRST",
    "FLOAT",
    "FOUND",
    "FREE",
    "FULLTEXTTABLE",
    "FUSION",
    "GENERAL",
    "GET",
    "GLOBAL",
    "GROUPING",
    "HOLD",
    "HOST",
    "HOUR",
    "IGNORE",
    "IMMEDIATE",
    "INDICATOR",
    "INITIALIZE",
    "INITIALLY",
    "INOUT",
    "INPUT",
    "INT",
    "INTEGER",
    "INTERSECTION",
    "INTERVAL",
    "ISOLATION",
    "ITERATE",
    "LANGUAGE",
    "LARGE",
    "LAST",
    "LATERAL",
    "LEADING",
    "LESS",
    "LEVEL",
    "LIKE_REGEX",
    "LIMIT",
    "LN",
    "LOCAL",
    "LOCALTIME",
    "LOCALTIMESTAMP",
    "LOCATOR",
    "MAP",
    "MATCH",
    "MEMBER",
    "METHOD",
    "MINUTE",
    "MOD",
    "MODIFIES",
    "MODIFY",
    "MODULE",
    "MONTH",
    "MULTISET",
    "NAMES",
    "NATURAL",
    "NCHAR",
    "NCLOB",
    "NEW",
    "NEXT",
    "NO",
    "NONE",
    "NORMALIZE",
    "NUMERIC",
    "OBJECT",
    "OCCURRENCES_REGEX",
    "OLD",
    "ONLY",
    "OPERATION",
    "ORDINALITY",
    "OUT",
    "OUTPUT",
    "OVERLAY",
    "PAD",
    "PARAMETER",
    "PARAMETERS",
    "PARTIAL",
    "PARTITION",
    "PATH",
    "PERCENT_RANK",
    "PERCENTILE_CONT",
    "PERCENTILE_DISC",
    "POSITION_REGEX",
    "POSTFIX",
    "PREFIX",
    "PREORDER",
    "PREPARE",
    "PRESERVE",
    "PRIOR",
    "PRIVILEGES",
    "RANGE",
    "READS",
    "REAL",
    "RECURSIVE",
    "REF",
    "REFERENCING",
    "REGR_AVGX",
    "REGR_AVGY",
    "REGR_COUNT",
    "REGR_INTERCEPT",
    "REGR_R2",
    "REGR_SLOPE",
    "REGR_SXX",
    "REGR_SXY",
    "REGR_SYY",
    "RELATIVE",
    "RELEASE",
    "RESULT",
    "RETURNING",
    "RETURNS",
    "ROLE",
    "ROLLUP",
    "ROUTINE",
    "ROW",
    "ROWS",
    "SAVEPOINT",
    "SCOPE",
    "SCROLL",
    "SEARCH",
    "SECOND",
    "SECTION",
    "SENSITIVE",
    "SEQUENCE",
    "SESSION",
    "SETS",
    "SIMILAR",
    "SIZE",
    "SMALLINT",
    "SPACE",
    "SPECIFIC",
    "SPECIFICTYPE",
    "SQL",
    "SQLEXCEPTION",
    "SQLSTATE",
    "SQLWARNING",
    "START",
    "STATE",
    "STATEMENT",
    "STATIC",
    "STDDEV_POP",
    "STDDEV_SAMP",
    "STRUCTURE",
    "SUBMULTISET",
    "SUBSTRING_REGEX",
    "SYMMETRIC",
    "SYSTEM",
    "TEMPORARY",
    "TERMINATE",
    "THAN",
    "TIME",
    "TIMESTAMP",
    "TIMEZONE_HOUR",
    "TIMEZONE_MINUTE",
    "TRAILING",
    "TRANSLATE_REGEX",
    "TRANSLATION",
    "TREAT",
    "TRUE",
    "UESCAPE",
    "UNDER",
    "UNKNOWN",
    "UNNEST",
    "USAGE",
    "USING",
    "VALUE",
    "VAR_POP",
    "VAR_SAMP",
    "VARCHAR",
    "VARIABLE",
    "WHENEVER",
    "WIDTH_BUCKET",
    "WINDOW",
    "WITHIN",
    "WITHOUT",
    "WORK",
    "WRAPPER",
    "WRITE",
    "XMLAGG",
    "XMLATTRIBUTES",
    "XMLBINARY",
    "XMLCAST",
    "XMLCOMMENT",
    "XMLCONCAT",
    "XMLDOCUMENT",
    "XMLELEMENT",
    "XMLEXISTS",
    "XMLFOREST",
    "XMLITERATE",
    "XMLNAMESPACES",
    "XMLPARSE",
    "XMLPI",
    "XMLQUERY",
    "XMLSERIALIZE",
    "XMLTABLE",
    "XMLTEXT",
    "XMLVALIDATE",
    "YEAR",
    "ZONE",
]

UNRESERVED_KEYWORDS = [
    "ABORT_AFTER_WAIT",
    "ABORT",
    "ABSENT",
    "ACCELERATED_DATABASE_RECOVERY",
    "ACCENT_SENSITIVITY",
    "ACCESS",
    "ADMINISTER",
    "ALGORITHM",
    "ALL_CONSTRAINTS",
    "ALL_ERRORMSGS",
    "ALL_INDEXES",
    "ALL_LEVELS",
    "ALLOW_CONNECTIONS",
    "ALLOW_ENCRYPTED_VALUE_MODIFICATIONS",
    "ALLOW_PAGE_LOCKS",
    "ALLOW_ROW_LOCKS",
    "ALLOW_SNAPSHOT_ISOLATION",
    "ALLOWED",
    "ALWAYS",
    "ANSI_DEFAULTS",
    "ANSI_NULL_DEFAULT",
    "ANSI_NULL_DFLT_OFF",
    "ANSI_NULL_DFLT_ON",
    "ANSI_NULLS",
    "ANSI_PADDING",
    "ANSI_WARNINGS",
    "APPEND_ONLY",
    "APPLICATION",
    "APPLY",
    "ARITHABORT",
    "ARITHIGNORE",
    "ASSEMBLY",
    "ATTACH",
    "ATTACH_REBUILD_LOG",
    "AUDIT",
    "AUTHENTICATE",
    "AUTO_CLOSE",
    "AUTO_CREATE_STATISTICS",
    "AUTO_CREATE_TABLE",
    "AUTO_SHRINK",
    "AUTO_UPDATE_STATISTICS",
    "AUTO_UPDATE_STATISTICS_ASYNC",
    "AUTO",
    "AVAILABILITY",
    "BACKUP_CLONEDB",
    "BACKUP_STORAGE_REDUNDANCY",
    "BATCHSIZE",
    "BERNOULLI",
    "BINDING",
    "BLOCK",
    "BLOCKERS",
    "BLOCKSIZE",
    "BULK_LOGGED",
    "BUFFERCOUNT",
    "CACHE",
    "CALLER",
    "CATALOG_COLLATION",
    "CATCH",
    "CHANGE",
    "CERTIFICATE",
    "CHANGE_TRACKING",
    "CHECK_CONSTRAINTS",
    "CHECK_EXPIRATION",
    "CHECK_POLICY",
    "CHECKALLOC",
    "CHECKSUM",
    "CHECKCATALOG",
    "CHECKCONSTRAINTS",
    "CHECKDB",
    "CHECKFILEGROUP",
    "CHECKIDENT",
    "CHECKTABLE",
    "CLASSIFICATION",
    "CLEANTABLE",
    "CLEAR",
    "CLONEDATABASE",
    "CODEPAGE",
    "COLLECTION",
    "COLUMN_ENCRYPTION_KEY",
    "COLUMNSTORE_ARCHIVE",
    "COLUMNSTORE",
    "COMMITTED",
    "COMPATIBILITY_LEVEL",
    "COMPRESS_ALL_ROW_GROUPS",
    "COMPRESSION_DELAY",
    "COMPRESSION",
    "CONCAT_NULL_YIELDS_NULL",
    "CONFIGURATION",
    "CONCAT",
    "CONTINUE_AFTER_ERROR",
    "CONNECTION_OPTIONS",
    "CONTAINED",
    "CONTENT",
    "CONTRACT",
    "COUNT_ROWS",
    "CONTAINMENT",
    "CONTEXT_INFO",
    "CONTROL",
    "COPY",
    "CREDENTIAL",
    "CRYPTOGRAPHIC",
    "CURSOR_CLOSE_ON_COMMIT",
    "CURSOR_DEFAULT",
    "DATA_COMPRESSION",
    "DATA_CONSISTENCY_CHECK",
    "DATA_DELETION",
    "DATA_PURITY",
    "DATA_SOURCE",
    "DATABASE_DEFAULT",
    "DATAFILETYPE",
    "DATASOURCE",
    "DATASPACE",
    "DATE_CORRELATION_OPTIMIZATION",
    "DATE_FORMAT",
    "DATEFIRST",
    "DATEFORMAT",
    "DAYS",
    "DB_CHAINING",
    "DBREINDEX",
    "DDL",
    "DEADLOCK_PRIORITY",
    "DECRYPTION",
    "DEFAULT_DATABASE",
    "DEFAULT_FULLTEXT_LANGUAGE",
    "DEFAULT_LANGUAGE",
    "DEFAULT_SCHEMA",
    "DELAY",
    "DELAYED_DURABILITY",
    "DELETED",
    "DELIMITEDTEXT",
    "DELTA",
    "DENSE_RANK",
    "DENSITY_VECTOR",
    "DIRECTORY_NAME",
    "DISABLE",
    "DISABLED",
    "DISTRIBUTION",
    "DISABLE_BROKER",
    "DOCUMENT",
    "DROP_EXISTING",
    "DROPCLEANBUFFERS",
    "DURABILITY",
    "EDGE",
    "EDITION",
    "ELASTIC_POOL",
    "ELEMENTS",
    "ELEVATE_ONLINE",
    "ELEVATE_RESUMABLE",
    "ENABLE",
    "ENABLE_BROKER",
    "ENCODING",
    "ENCRYPTED",
    "ENCRYPTION_TYPE",
    "ENCRYPTION",
    "ENDPOINT",
    "ERROR_BROKER_CONVERSATIONS",
    "ERRORFILE_CREDENTIAL",
    "ERRORFILE_DATA_SOURCE",
    "ERRORFILE_LOCATION",
    "ERRORFILE",
    "ESCAPE_CHAR",
    "ESTIMATEONLY",
    "EVENT",
    "EXPAND",
    "EXPLAIN",
    "EXPLICIT",
    "EXTENDED_LOGICAL_CHECKS",
    "EXTERNALPUSHDOWN",
    "FAILOVER",
    "FAIL_UNSUPPORTED",
    "FAST",
    "FAST_FORWARD",
    "FIELD_TERMINATOR",
    "FIELDQUOTE",
    "FIELDTERMINATOR",
    "FILE_FORMAT",
    "FILE_TYPE",
    "FILEGROUP",
    "FILEGROWTH",
    "FILENAME",
    "FILESTREAM_ON",
    "FILESTREAM",
    "FILETABLE_COLLATE_FILENAME",
    "FILETABLE_DIRECTORY",
    "FILETABLE_FULLPATH_UNIQUE_CONSTRAINT_NAME",
    "FILETABLE_PRIMARY_KEY_CONSTRAINT_NAME",
    "FILETABLE_STREAMID_UNIQUE_CONSTRAINT_NAME",
    "FILTER_COLUMN",
    "FILTER_PREDICATE",
    "FIPS_FLAGGER",
    "FIRE_TRIGGERS",
    "FIRST_ROW",
    "FIRSTROW",
    "FLUSHAUTHCACHE",
    "FMTONLY",
    "FOLLOWING",
    "FORCE",
    "FORCED",
    "FORCE_FAILOVER_ALLOW_DATA_LOSS",
    "FORCEPLAN",
    "FORCESCAN",
    "FORCESEEK",
    "FORMAT_OPTIONS",
    "FORMAT_TYPE",
    "FORMAT",
    "FORMATFILE_DATA_SOURCE",
    "FORMATFILE",
    "FORWARD_ONLY",
    "FREEPROCCACHE",
    "FREESESSIONCACHE",
    "FREESYSTEMCACHE",
    "FULLSCAN",
    "FULLTEXT",
    "GB",
    "GENERATED",
    "GEO",
    "GEOZONE",
    "HASH",
    "HEAP",
    "HEADER_ROW",
    "HELP",
    "HIDDEN",
    "HIGH",
    "HINT",
    "HISTOGRAM",
    "HISTORY_RETENTION_PERIOD",
    "HISTORY_TABLE",
    "HONOR_BROKER_PRIORITY",
    "IGNORE_CONSTRAINTS",
    "IGNORE_DUP_KEY",
    "IGNORE_NONCLUSTERED_COLUMNSTORE_INDEX",
    "IGNORE_TRIGGERS",
    "IMPERSONATE",
    "IMPLICIT_TRANSACTIONS",
    "INBOUND",
    "INCLUDE_NULL_VALUES",
    "INCLUDE",
    "INCREMENT",
    "INCREMENTAL",
    "INDEXDEFRAG",
    "INFINITE",
    "INLINE",
    "INPUTBUFFER",
    "INSENSITIVE",
    "INSERTED",
    "INSTEAD",
    "INTEGRATION",
    "IO",
    "ISJSON",
    "JSON",
    "JSON_ARRAY",
    "JSON_ARRAYAGG",
    "JSON_MODIFY",
    "JSON_OBJECT",
    "JSON_OBJECTAGG",
    "JSON_PATH_EXISTS",
    "JSON_QUERY",
    "JSON_VALUE",
    "KB",
    "KEEP",
    "KEEPDEFAULTS",
    "KEEPFIXED",
    "KEEPIDENTITY",
    "KEEPNULLS",
    "KEYSET",
    "KILOBYTES_PER_BATCH",
    "LABEL",
    "LASTROW",
    "LEDGER_VIEW",
    "LEDGER",
    "LEGACY_CARDINALITY_ESTIMATION",
    "LIBRARY",
    "LINKED",
    "LIST",
    "LOB_COMPACTION",
    "LOCATION",
    "LOCK_TIMEOUT",
    "LOG",
    "LOGIN",
    "LOOP",
    "LOW",
    "MANUAL",
    "MANUAL_CUTOVER",
    "MASK",
    "MASKED",
    "MASTER",
    "MATCHED",
    "MAX_DURATION",
    "MAX_GRANT_PERCENT",
    "MAX",
    "MARK_IN_USE_FOR_REMOVAL",
    "MAXDOP",
    "MAXERRORS",
    "MAXRECURSION",
    "MAXSIZE",
    "MAXTRANSFERSIZE",
    "MAXVALUE",
    "MB",
    "MEMORY_OPTIMIZED",
    "MEMORY_OPTIMIZED_DATA",
    "MIGRATION_STATE",
    "MIN_GRANT_PERCENT",
    "MINUTES",
    "MINVALUE",
    "MODEL",
    "MOVE",
    "MONTHS",
    "MULTI_USER",
    "MEDIANAME",
    "MEDIAPASSWORD",
    "MESSAGE",
    "MUST_CHANGE",
    "NAME",
    "NATIVE_COMPILATION",
    "NESTED_TRIGGERS",
    "NEWNAME",
    "NEW_BROKER",
    "NO_INFOMSGS",
    "NO_QUERYSTORE",
    "NO_STATISTICS",
    "NO_PERFORMANCE_SPOOL",
    "NO_CHECKSUM",
    "NO_TRUNCATE",
    "NO_WAIT",
    "NOCOUNT",
    "NORECOVERY",
    "NOREWIND",
    "NOUNLOAD",
    "NODE",
    "NOEXEC",
    "NOEXPAND",
    "NOLOCK",
    "NON_TRANSACTED_ACCESS",
    "NORMAL",
    "NOTIFICATION",
    "NOTIFICATIONS",
    "NOWAIT",
    "NTILE",
    "NULLS",
    "NUMERIC_ROUNDABORT",
    "OFFSET",
    "OLD_PASSWORD",
    "ONLINE",
    "OPENJSON",
    "OPENTRAN",
    "OPERATION_TYPE_COLUMN_NAME",
    "OPERATION_TYPE_DESC_COLUMN_NAME",
    "OPERATIONS",
    "OPTIMISTIC",
    "OPTIMIZE_FOR_SEQUENTIAL_KEY",
    "OPTIMIZE",
    "ORC",
    "OUTBOUND",
    "OUTPUTBUFFER",
    "OVERRIDE",
    "OWNER",
    "OWNERSHIP",
    "PAD_INDEX",
    "PAGE",
    "PAGE_VERIFY",
    "PAGLOCK",
    "PARAMETERIZATION",
    "PARQUET",
    "PARSEONLY",
    "PARSER_VERSION",
    "PARTITIONS",
    "PASSWORD",
    "PAUSE",
    "PAUSED",
    "PERCENTAGE",
    "PERFORM_CUTOVER",
    "PERIOD",
    "PERSISTED",
    "PERSISTENT_LOG_BUFFER",
    "PERSISTENT_VERSION_STORE_FILEGROUP",
    "PHYSICAL_ONLY",
    "POLICY",
    "POPULATION",
    "PRECEDING",
    "PREDICATE",
    "PROCCACHE",
    "PROCEDURE_CACHE",
    "PREDICT",
    "PROFILE",
    "PROPERTY",
    "PROVIDER",
    "PUSHDOWN",
    "QUERY",
    "QUERY_GOVERNOR_COST_LIMIT",
    "QUERY_STORE",
    "QUERYTRACEON",
    "QUEUE",
    "QUOTED_IDENTIFIER",
    "R",
    "RANDOMIZED",
    "RANK",
    "RAW",
    "RCFILE",
    "READCOMMITTED",
    "READCOMMITTEDLOCK",
    "READONLY",
    "READPAST",
    "READ_COMMITTED_SNAPSHOT",
    "READ_ONLY",
    "READ_WRITE",
    "READ_WRITE_FILEGROUPS",
    "READUNCOMMITTED",
    "REBUILD",
    "RECEIVE",
    "RECOMPILE",
    "REMOTE",
    "RESTART",
    "RECOVERY",
    "RECURSIVE_TRIGGERS",
    "REGENERATE",
    "REJECT_SAMPLE_VALUE",
    "REJECT_TYPE",
    "REJECT_VALUE",
    "REJECTED_ROW_LOCATION",
    "REMOTE_DATA_ARCHIVE",
    "REMOTE_PROC_TRANSACTIONS",
    "REMOVE",
    "RENAME",
    "REORGANIZE",
    "REPEATABLE",
    "REPEATABLEREAD",
    "REPLACE",
    "REPLICATE",
    "RESAMPLE",
    "RESEED",
    "RESOURCES",
    "REWIND",
    "RESPECT",
    "RESULT_SET_CACHING",
    "RESUMABLE",
    "RESTRICTED_USER",
    "RESUME",
    "RETENTION_PERIOD",
    "ROBUST",
    "ROOT",
    "ROUND_ROBIN",
    "ROUTE",
    "ROW_NUMBER",
    "ROWLOCK",
    "ROWS_PER_BATCH",
    "ROWSET_OPTIONS",
    "ROWTERMINATOR",
    "S",
    "SCALAR",
    "SCALEOUTEXECUTION",
    "SCHEMA_AND_DATA",
    "SCHEMA_ONLY",
    "SCHEMABINDING",
    "SCHEME",
    "SCOPED",
    "SCRIPT",
    "SCROLL_LOCKS",
    "SECONDARY",
    "SECONDS",
    "SECRET",
    "SECURABLES",
    "SECURITY",
    "SELF",
    "SEND",
    "SENSITIVITY",
    "SEQUENCE_NUMBER_COLUMN_NAME",
    "SEQUENCE_NUMBER",
    "SERDE_METHOD",
    "SERIALIZABLE",
    "SERVER",
    "SERVICE",
    "SERVICEBROKER",
    "SERVICE_OBJECTIVE",
    "SETERROR",
    "SETTINGS",
    "SETVAR",
    "SHOWPLAN",
    "SHOW_STATISTICS",
    "SHOWCONTIG",
    "SHOWPLAN_ALL",
    "SHOWPLAN_TEXT",
    "SHOWPLAN_XML",
    "SHRINKDATABASE",
    "SHRINKFILE",
    "SID",
    "SIMPLE",
    "SINGLE_BLOB",
    "SINGLE_CLOB",
    "SINGLE_NCLOB",
    "SNAPSHOT",
    "SORT_IN_TEMPDB",
    "SOURCE",
    "SPARSE",
    "SPATIAL_WINDOW_MAX_CELLS",
    "SPLIT",
    "STANDBY",
    "SQLPERF",
    "STAT_HEADER",
    "STATS",
    "STATISTICAL_SEMANTICS",
    "STATISTICS_INCREMENTAL",
    "STATISTICS_NORECOMPUTE",
    "STOP",
    "STOPLIST",
    "STOP_ON_ERROR",
    "STRING_AGG",
    "STRING_DELIMITER",
    "SUBSCRIBE",
    "SWITCH",
    "SYNONYM",
    "SYSTEM_TIME",
    "SYSTEM_VERSIONING",
    "TB",
    "TABLERESULTS",
    "TABLOCK",
    "TABLOCKX",
    "TAPE",
    "TAKE",
    "TARGET",
    "TARGET_RECOVERY_TIME",
    "TEXTIMAGE_ON",
    "THROW",
    "TIES",
    "TIMEOUT",
    "TRACE",
    "TRACEOFF",
    "TRACEON",
    "TRACESTATUS",
    "TRACKING",
    "TRANSACTION_ID_COLUMN_NAME",
    "TRANSACTION_ID",
    "TRANSFORM_NOISE_WORDS",
    "TRUNCATE_TARGET",
    "TRUSTWORTHY",
    "TRY",
    "TYPE",
    "TYPE_WARNING",
    "TWO_DIGIT_YEAR_CUTOFF",
    "UNBOUNDED",
    "UNCOMMITTED",
    "UNDEFINED",
    "UNLIMITED",
    "UNMASK",
    "UNLOAD",
    "UNSAFE",
    "UPDATEUSAGE",
    "UPDLOCK",
    "URL",
    "USE_TYPE_DEFAULT",
    "USED",
    "USER_DB",
    "USEROPTIONS",
    "VIEW_METADATA",
    "WAIT_AT_LOW_PRIORITY",
    "WHEN_SUPPORTED",
    "WEEK",
    "WEEKS",
    "WINDOWS",
    "WITHOUT_ARRAY_WRAPPER",
    "XACT_ABORT",
    "XLOCK",
    "XML_COMPRESSION",
    "XML",
    "XMLDATA",
    "XMLSCHEMA",
    "XSINIL",
    "YEARS",
    # Data type keywords
    # https://learn.microsoft.com/en-us/sql/t-sql/data-types/data-types-transact-sql
    # Exact numeric types
    "TINYINT",
    "BIGINT",
    "MONEY",
    "SMALLMONEY",
    # Approximate numeric types
    # Date and time types
    "SMALLDATETIME",
    "DATETIME",
    "DATETIME2",
    "DATETIMEOFFSET",
    # Character string types
    "TEXT",
    # Unicode character string types
    "NVARCHAR",
    "NTEXT",
    # Binary string types
    "VARBINARY",
    "IMAGE",
    # Other data types
    "SQL_VARIANT",
    "ROWVERSION",
    "UNIQUEIDENTIFIER",
    "GEOGRAPHY",
    "GEOMETRY",
    "HIERARCHYID",
    "VECTOR",
    "VERIFY_CLONEDB",
    "WAREHOUSE",
]

# DataType Methods
# Unlike normal keywords datatype methods are case sensitive
# grouped by datatype and geometry/spatial data classification
DATATYPE_METHODS = [
    # xml methods
    # https://learn.microsoft.com/en-us/sql/t-sql/xml/xml-data-type-methods
    "exist",
    "modify",
    "nodes",
    "query",
    "value",
    # hierarchyid methods
    # https://learn.microsoft.com/en-us/sql/t-sql/data-types/hierarchyid-data-type-method-reference
    "GetAncestor",
    "GetLevel",
    "GetReparentedValue",
    "GetRoot",
    "IsDescendantOf",
    "Parse",
    "ToString",
    # geometry (spatial data) methods
    # https://learn.microsoft.com/en-us/sql/t-sql/spatial-geometry/ogc-methods-on-geometry-instances
    # Shape properties
    "STArea",
    "STLength",
    "STBoundary",
    "STCentroid",
    "STEnvelope",
    "STConvexHull",
    # Geometry representation
    "STAsBinary",
    "STAsText",
    # Geometry type information
    "STGeometryType",
    "STDimension",
    "STSrid",
    # Point and curve access
    "STStartPoint",
    "STEndPoint",
    "STPointN",
    "STPointOnSurface",
    "STCurveN",
    "STCurveToLine",
    "STX",
    "STY",
    # Polygon ring access
    "STExteriorRing",
    "STInteriorRingN",
    "STNumInteriorRing",
    # Collection access
    "STGeometryN",
    "STNumGeometries",
    "STNumPoints",
    "STNumCurves",
    # Spatial relationship tests
    "STContains",
    "STCrosses",
    "STDisjoint",
    "STEquals",
    "STIntersects",
    "STOverlaps",
    "STRelate",
    "STTouches",
    "STWithin",
    "STDistance",
    # Spatial operations
    "STBuffer",
    "STDifference",
    "STIntersection",
    "STSymDifference",
    "STUnion",
    # Validity tests
    "STIsClosed",
    "STIsEmpty",
    "STIsRing",
    "STIsSimple",
    "STIsValid",
    # geography methods share the same ST-prefixed names as geometry methods
    # and are already covered by the geometry section above.
    # https://learn.microsoft.com/en-us/sql/t-sql/spatial-geography/ogc-methods-on-geography-instances
]



"""The KeywordSegment class."""

from typing import Callable, Optional, Union

from sqlfluff.core.parser.markers import PositionMarker
from sqlfluff.core.parser.segments.base import SourceFix
from sqlfluff.core.parser.segments.common import WordSegment


class KeywordSegment(WordSegment):
    """A segment used for matching single words.

    We rename the segment class here so that descendants of
    _ProtoKeywordSegment can use the same functionality
    but don't end up being labelled as a `keyword` later.
    """

    type = "keyword"

    def __init__(
        self,
        raw: Optional[str] = None,
        pos_marker: Optional[PositionMarker] = None,
        instance_types: tuple[str, ...] = (),
        source_fixes: Optional[list[SourceFix]] = None,
        trim_chars: Optional[tuple[str, ...]] = None,
        quoted_value: Optional[tuple[str, Union[int, str]]] = None,
        escape_replacements: Optional[list[tuple[str, str]]] = None,
        casefold: Optional[Callable[[str], str]] = None,
    ):
        """If no other name is provided we extrapolate it from the raw."""
        super().__init__(
            raw=raw,
            pos_marker=pos_marker,
            instance_types=instance_types,
            source_fixes=source_fixes,
            trim_chars=trim_chars,
            quoted_value=quoted_value,
            escape_replacements=escape_replacements,
            casefold=casefold,
        )

    def edit(
        self, raw: Optional[str] = None, source_fixes: Optional[list[SourceFix]] = None
    ) -> "KeywordSegment":
        """Create a new segment, with exactly the same position but different content.

        Returns:
            A copy of this object with new contents.

        Used mostly by fixes.

        NOTE: This *doesn't* copy the uuid. The edited segment is a new segment.

        """
        return self.__class__(
            raw=raw or self.raw,
            pos_marker=self.pos_marker,
            instance_types=self.instance_types,
            source_fixes=source_fixes or self.source_fixes,
        )


class LiteralKeywordSegment(KeywordSegment):
    """A keyword style literal segment.

    This should be used for things like NULL, NAN, TRUE & FALSE.

    Defined here for type inheritance.
    """

    type = "literal"



"""Helper methods for other classes.

This module should have no dependencies on other parts of
the SQLFluff package and each should also be free of dependencies
on each other. This is to ensure each of these methods can be used
anywhere within the project without fear of dependency issues.

Methods are organised by the datatype they are designed to help with.
"""
